/*
 * Component: utils
 *
 * Description:
 * 	Header file for helper functions related to advanced resource handling.
 * 	Advanced resource handling means:
 * 	Repository is a directory with a configuration file (optionally). 
 * 	The repository contains at least one mount.
 * 		A mount is a subdirectory in the repository.
 * 		Each mount is meant to be a different partition/share.
 * 		Each mount must contain a configuration file called repositorypart.cfg. 
 * 		The repositorypart.cfg may be empty, it's existence indicates that the partition/share is mounted.
 * 	A repository with no config file is supposed to contain the hashed directories in it's root 
 * 	(the repository itself is it's only mount).
 *	If a config file (called "repository.cfg") is present, it may contain the following data:
 * 	legacyDir -> a place where files and directories exist without hashing. 
 * 		When a resource is found to be in the legacyDir it is to be transferred to the repository
 *	mountPoints -> a backslash ('\') separated list of subdirectories/mounts in the repository. 
 *	remindLimit -> if the space usage on each available mount points exceeds this threshold the system displays 
 * 		a warning message to all users in the admin group (default value is 80%)
 *	warnLimit -> if the space usage on each available mount points exceeds this threshold the system displays 
 * 		a warning message to all users in the admin group and resets it's acknowlidgements once a day 
 * 		(default value is 90%)
 *	switchLimit -> when creating new resources the first mount is used that does not exceed this threshold 
 *  	(default value is 95%). If all mounts exceed this threshold but there is at least one below fullLimit
 * 		the mount with the most space (in percentage) will be used. If all mounts are above fullLimit, the 
 * 		repository will not allow to create new resources 
 * 	fullLimit -> new resources may be inserted into mounts below this threshold only. Defaults to 97%
 * 	cleanLimit -> when moving resources from a filled-up mount, move resources until space usage drops below 
 * 		this limit (default value is 98%)
 * 	moveLimit -> if the space usage exceeds this threshold on a mount point, the system moves the referenced
 *		resource to the current mountPoint (default value is 99%)
 * 	nomoveTime -> in days, only resources older than this can be moved/deleted
 *  isClearable -> if true, the oldest resources will be deleted instead of moved when the mount is above moveLimit
 * 
 *  - the repositorypart.cfg file can override the following data:
 *   switchLimit, fullLimit, cleanLimit and moveLimit 
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef __IM_DIR_HANDLER_H_INCLUDED__
#define __IM_DIR_HANDLER_H_INCLUDED__

#include <jni.h>

#include <utime.h>

#include <unistd.h>

#include <string>
#include <list>
using namespace std;

typedef enum {
	REPOSITORY_STATUS_OK = 0, 					// everything is perfect
	REPOSITORY_STATUS_REMIND = 1,				// the only mountpoint below switchLimit is above remind limit 
														// or no mount point below switchLimit
	REPOSITORY_STATUS_WARN = 2,					// the only mountpoint below switchLimit is above warn limit 
															// or no mount point below switchLimit
	REPOSITORY_STATUS_WARN2 = 4,				// the only mountpoint below switchLimit is above warn limit 
														// or no mount point below switchLimit
	REPOSITORY_STATUS_WARN3 = 8,				// the only mountpoint below switchLimit is above warn limit 
														// or no mount point below switchLimit
	REPOSITORY_STATUS_SWITCH = 16,			// there is no mount point below switchLimit
	REPOSITORY_STATUS_CANNOTMOVE = 32,  // there is no mountpoint to move data 
	REPOSITORY_STATUS_FULL = 64,				// there is no mount point below fullLimit 
														// (the mount does not allow new resources)
	REPOSITORY_STATUS_INCOMPLETE = 128,	// at least one of the configured mount points is not mounted
	REPOSITORY_STATUS_MISSING = 256			// the entire repository is missing
} REPOSITORY_STATUS;

typedef enum {
	MOUNT_STATUS_OK = 0,				// everything is perfect
	MOUNT_STATUS_REMIND = 1,		// this mountpoint is above remind limit
	MOUNT_STATUS_WARN = 2,			// this mountpoint is above warn limit
	MOUNT_STATUS_WARN2 = 4,			// this mountpoint is above warn limit
	MOUNT_STATUS_WARN3 = 8,			// this mountpoint is above warn limit
	MOUNT_STATUS_SWITCH = 16,		// this mountpoint is above switch limit
	MOUNT_STATUS_FULL = 32,			// this mountpoint is above full limit
	MOUNT_STATUS_CLEAN = 64,		// this mountpoint is above clean limit
	MOUNT_STATUS_MOVE = 128,			// this mountpoint is above move limit
	MOUNT_STATUS_UNMOUNTED = 256	// this mountpoint is not mounted
} MOUNT_STATUS;

typedef enum {
	RESOURCE_STATUS_OK = 0,						// the resource is available and non-conflicting
	RESOURCE_STATUS_DIRTY = 1,				// the resource is available and may be conflicting 
													// (missing mount may contain it too)
	RESOURCE_STATUS_CONFLICTING = 2,	// the resource is available at two locations at least and 
												// it can not be automatically merged
	RESOURCE_STATUS_MISSING = 4,			// the resource could not be located
	RESOURCE_STATUS_ERRONEOUS = 8,		// the resource can not be accessed (IO error for example)
	RESOURCE_STATUS_NOLOCK = 16				// the resource can not be locked
} RESOURCE_STATUS;

#ifdef __cplusplus
extern "C" {
#endif

#define isDEBUGREP  true

#define DEBUGREP(...) if (isDEBUGREP) { rh_logger.debugLog(__VA_ARGS__); }

#define ERRORREP(...) { rh_logger.errorLog(__VA_ARGS__); }

#define INFOREP(...) { rh_logger.infoLog(__VA_ARGS__); }

#define OPERINFOREP(...) { rh_logger.operInfoLog(__VA_ARGS__); }

// ---------------------------------- C interface -------------------------------------------------------

/*
 * Creates a Directory in a given repository
 * 
 * @param repository	where to put the new Directory
 * @param subDir		the subdirectory to be created.
 * 		if the repository is located in "/aaa", the current mount point is "bbb" and the subDir is "xyz"
 * 		the created directory will look like "/aaa/bbb/x1/x2/x3/xyz" where x1, x2 and x3 are hash
 * 		values generated from "xyz"
 * @param stat			int pointer that (if present) will store the resource status if it already exists
 * @param mustBeNew	bool indicating if the caller is certain that the resource does not exist yet 
 * 		(don't mark it as dirty if the repository is incomplete)
 * @return	0 if no error occured
 */
char* createDirectory_( const char* repository, const char* subDir, int* stat = NULL, bool mustBeNew = false );

/*
 * Creates the path that will contain a file in a given repository
 * 
 * @param repository	where to put the new File
 * @param fileName	the file for which the path should be created.
 * 		if the repository is located in "/aaa", the current mount point is "bbb" and the file is "xyz"
 * 		the created directory will look like "/aaa/bbb/x1/x2/x3" where x1, x2 and x3 are hash 
 * 		values generated from "xyz"
 * @param stat			int pointer that (if present) will store the resource status if it already exists
 * @param mustBeNew	bool indicating if the caller is certain that the resource does not exist yet 
 * 		(don't mark it as dirty if the repository is incomplete)
 * @return	0 if no error occured
 */
char* createLocation_( const char* repository, const char* fileName, int* stat = NULL, bool mustBeNew = false );

/*
 * Returns the path that contains a file/directory in a given repository
 * 
 * @param repository	where to locate the resource
 * @param resource	the file/directory that should be located.
 * @return	absolute path name of the requested resource, or empty string if not found
 */
char* getLocation_( const char* repository, const char* resource, int* stat = NULL );

/*
 * this method checks the status of a given repository and if it does not match with the result of the previous 
 * check do certain tasks that the change justifies
 * @param repository	which repository to check
 * @return	int value describing the status of the repository (like REPOSITORY_STATUS_OK)
 */
int getRepositoryStatus_( const char* repository );

/*
 * this method checks the status of a given resource in a given repository
 * @param repository	in which repository to check
 * @param resource	the file/directory that should be checked.
 * @return	int value describing the status of the resource (like RESOURCE_STATUS_OK)
 */
int checkResourceStatus_( const char* repository, const char* resource );

/*
 * this method removes a resource (recursively if it is a directory) and hashed dirs that became empty
 * by this action
 * if the resource is dirty or not found, remove will not do anything except returning status
 * @param repository	from which repository to remove
 * @param resource	the file/directory that should be removed.
 * @return resource status ( 0 if no error occured )
 */
int removeResource_( const char* repository, const char* resource );

/*
 * this method lists the files stored in a given repository 
 * @param repository	where to list the files
 * @param count				long* that will hold the size of the returned array
 * @param sort				indicates whether the result should be sorted
 * @return	string array with the filenames (without path, like "stuff.txt")
 */
char** listFiles_( const char* repository, long* count, bool sort = false );

/*
 * this method lists the directories stored in a given repository 
 * @param repository	where to list the directories
 * @param count				long* that will hold the size of the returned array
 * @param sort				indicates whether the result should be sorted
 * @return	string array with the directorynames (like "xyz/" or "xyz/abc/")
 */
char** listDirs_( const char* repository, long* count, bool sort = false );

/*
 * this method is for testing
 * clears the repository
 * @param repository	
 */
void makeEmpty_( const char* repository );

/*
 * this method is for testing
 * creates a directory to the legacyDir, if possible
 * @param repository	
 * @param directory to create
 * @return name of the created directory
 */
char* createToLegacy_( const char* repository, const char* resource );

/*
 * this method is for testing
 * mounts/unmounts a drive randomly
 * @param repository	
 */
void rndMount_( const char* repository );

/*
 * this method is for testing
 * mounts/unmounts a given drive
 * @param repository
 * @return -1 if the mount doesn't exist, 0 otherwise	
 */
int MountUnmount_( const char* repository, int num, bool mount = true );

/*
 * this method is for testing
 * mounts all mountpoints
 * @param repository	
 */
void giveBackMounts_( const char* repository );

/*
 * this method is for testing
 * after checkOverload it gives back the names of deleted resources
 * @param repository	
 * @return the pointer to the set with the names of resources 
 */
list< string >* getDeletedDirs_( const char* repository );

/*
 * this method converts a repository name and a resource name into a unique identifier
 * that can be used to lock the resource;
 * @param	repository	name of the repository
 * @param	resource		name of the resource
 * @return	unique identifier that can be used with IMStlSemaphore
 */
char* toResourceID( const char* repository, const char* resource );

/*
 * this method returns the last access time of a given resource
 * @param	resource	absolute pathname of the resource to be checked
 * @param	resStat		if the stat call for the resource has alrady been made, it's result may be presented here to avoid an extra stat call
 * @return	the last access time as stat reported it. 
 */
time_t lastAccess( const char* resource, struct stat* resStat_ = NULL );

/*
 * this method returns the last modification time of a given file
 * @param	file	absolute pathname of the resource to be checked
 * @return	the last access time as stat reported it.
 */
time_t lastModified( const char* file );

/*
 * this is the real constructor
 * also used for re-configuration
 * @param  the name of the repository
 */
void setup_( const char* repository );

/*
 * this method checks the mounts, and if a mount is above move level, tries to move or delete 
 * files and directories until it goes below clean level
 * @param repository	where to check
 */
void checkOverload_( const char* repository, bool verbose = false );

/*
 * this method tries to completely empty a mount in a repository by moving data safely to other mounts
 * @param repository	 	the repository
 * @param mountSubDir	the subdirectory in the repository that should be emptied
 */
int freeMount_( const char* repository, const char* mountSubDir, bool verbose = false );

/*
 * this method calculates the size of the given directory 
 * !assumed to be the root of a mount, or a legacyDir 
 * uses the .count file in the directory, if exists
 * also uses the temporary .countres file there
 * @param dir  to count
 * @param legacy  if true, treat it as a legacyDir
 * @param recalcOldStudiesNum if not 0, this number of the oldest accessed studies will be forced to recalc ('du')
 * @return  the size of the directory in bytes
 */
long long int countSize( const char* dir, unsigned int recalcOldStudiesNum = 0, bool legacy = false );

/**
 * If the resource is not on a 'fast' mount, try to find a 'fast' mount (whcich hasn't reached switch or full limit)
 *   and move the resource there
 *   access time will not change
 *
 * @param repository	 	the repository
 * @param	resource	the name of the resource to move
 * @return	the new location of the resource if it was in fact moved, NULL otherwise
 */
char * unAge_( const char* repository, const char* resource );

bool getTimeProperties_( const char* repository, const char* resource, utimbuf* props );

bool setTimeProperties_( const char* repository, const char* resource, utimbuf* props );

void sendLog( int level, const char* message, ... );

// ---------------------------------- JNI interface -----------------------------------------------------

/*
 * Class:     imagemed_util_RepositoryHandler
 * Method:    createDirectoryNative
 * Signature: (Ljava/lang/String;Ljava/lang/String;Z)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_imagemed_util_RepositoryHandler_createDirectoryNative( JNIEnv *, jobject, jstring, jstring,
		jboolean );

/*
 * Class:     imagemed_util_RepositoryHandler
 * Method:    createLocationNative
 * Signature: (Ljava/lang/String;Ljava/lang/String;Z)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_imagemed_util_RepositoryHandler_createLocationNative( JNIEnv *, jobject, jstring, jstring,
		jboolean );

/*
 * Class:     imagemed_util_RepositoryHandler
 * Method:    getLocationNative
 * Signature: (Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_imagemed_util_RepositoryHandler_getLocationNative( JNIEnv *, jobject, jstring, jstring );

/*
 * Class:     imagemed_util_RepositoryHandler
 * Method:    getRepositoryStatusNative
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_imagemed_util_RepositoryHandler_getRepositoryStatusNative( JNIEnv *, jobject, jstring );

/*
 * Class:     imagemed_util_RepositoryHandler
 * Method:    checkResourceStatusNative
 * Signature: (Ljava/lang/String;Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_imagemed_util_RepositoryHandler_checkResourceStatusNative( JNIEnv *, jobject, jstring, jstring );

/*
 * Class:     imagemed_util_RepositoryHandler
 * Method:    removeResourceNative
 * Signature: (Ljava/lang/String;Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_imagemed_util_RepositoryHandler_removeResourceNative( JNIEnv *, jobject, jstring, jstring );

/*
 * Class:     imagemed_util_RepositoryHandler
 * Method:    listFilesNative
 * Signature: (Ljava/lang/String;Z)[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_imagemed_util_RepositoryHandler_listFilesNative( JNIEnv *, jobject, jstring, jboolean );

/*
 * Class:     imagemed_util_RepositoryHandler
 * Method:    listDirsNative
 * Signature: (Ljava/lang/String;Z)[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_imagemed_util_RepositoryHandler_listDirsNative( JNIEnv *, jobject, jstring, jboolean );

/*
 * Class:     imagemed_util_RepositoryHandler
 * Method:    unAgeNative
 * Signature: (Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_imagemed_util_RepositoryHandler_unAgeNative( JNIEnv *, jobject, jstring, jstring );

#ifdef __cplusplus
}
#endif

#endif // __IM_DIR_HANDLER_H_INCLUDED__
